"""Define the RecordSet Class."""
from .exceptions import UnknownHeadingError, NoScalarDataError


class RecordSet:
    """
    The RecordSet Object.

    This is the object returned by the ConnectionManager from the recordset
    method. Saves the output of the SQL query once, but allows several ways
    to access it.

    This object is initialised using the normal __init__ method, as there
    isn't much use for an instance of a totally empty recordset. If we need it
    in the future, we can move the __init__ method to a class method.
    """

    def __init__(self, headings, data):
        """
        Initialise this object with data and headings.

        <data> can be None, representing no rows of data for the given
        headings.
        """
        self.data = data or None
        self.headings = headings

    def __bool__(self):
        return self.data is not None

    def __iter__(self):
        if self:
            return iter(self.data)
        return iter(())

    def dict_gen(self):
        """Iterate over records as dictionaries."""
        if self:
            for record in self.data or []:
                yield {h: v for h, v in zip(self.headings, record)}
        return []

    def as_dict(self):
        """
        Return the records as a list of dicts.

        Creates a copy of the data, so isn't very efficient for larger
        data sets.
        """
        if self:
            return [
                {h: v for h, v in zip(self.headings, record)}
                for record in self.data or []
            ]
        return []

    def column(self, heading):
        """Return a list of data for a particular heading."""
        try:
            heading_index = self.headings.index(heading)
        except ValueError as e:
            raise UnknownHeadingError(heading) from e

        return [record[heading_index] for record in self.data or []]


class RecordScalar:
    """
    The RecordScalar object assumes that a single value is being returned.

    For example, you would use this when you are trying to find 1 data point
    for a single identity. The same data passed to a RecordSet object is
    passed to this class, however if multiple columns or rows are passed,
    only the top left value is kept.

    Note, we have to distinguish between a NULL value returned, and no rows
    returned. __bool__ is therefore "was a record returned" by the query, and
    then you use self.datum truthiness as per normal.
    """

    def __init__(self, headings, data):
        # Only keep the first heading
        self.heading = headings[0]

        # Only keep None or the first value of data.
        if data:
            self._datum = [data[0][0]]
        else:
            self._datum = None

    def __bool__(self):
        # was a record returned
        return bool(self._datum)

    @property
    def datum(self):
        if not self:
            raise NoScalarDataError()
        return self._datum[0]

    def sdatum(self, default=None):
        """
        Return the datum with a default value if the record doesn't exist.

        sdatum stands for Safe Datum, providing a default value instead of 
        raising an error if the record doesn't exist.
        """
        if not self:
            return default
        return self._datum[0]


class Record:
    """
    The Record object assumes only a single record is being returned.

    If multiple rows are returned, only the first is kept.
    """

    def __init__(self, headings, data):
        self.data = data[0] if data else None

        self.headings = headings
        if self.data:
            self._dict = {k: v for k, v in zip(self.headings, self.data)}
        else:
            self._dict = {}

    def __getitem__(self, heading):
        try:
            return self._dict[heading]
        except KeyError as e:
            raise UnknownHeadingError(heading) from e

    @property
    def as_dict(self):
        return self._dict

    def __bool__(self):
        return self.data is not None



from sqlalchemy import text, VARCHAR, bindparam


def bind_sql(sql, params):
    """
    Bind parameters to a SQL query.

    Useful if the library that is executing your query, for exaple pandas,
    doesn't like named parameters. bind_sql will return a sql query with the
    parameters bound to the query.

    Args:
        sql: The SQL query to bind parameters to
        params: The named parameters to bind to the query
    """
    bound_sql = text(sql)  # convert to the useful sqlalchemy text

    if params:
        for key, value in params.items():
            # If the type of the value of the parameter is str, then we use
            # the VARCHAR object with no maximum, else use the default Type
            type_ = VARCHAR(None) if isinstance(value, str) else None
            param = bindparam(key=key, value=value, type_=type_)
            bound_sql = bound_sql.bindparams(param)

    return bound_sql



"""Define exceptions for the RecordSet Object."""


class UnknownHeadingError(Exception):
    def __init__(self, heading):
        msg = "Unknown heading: [{}]".format(heading)
        super().__init__(msg)
        self.message = msg


class NoScalarDataError(Exception):
    def __init__(self):
        msg = "No datum was returned"
        super().__init__(msg)
        self.message = msg



"""Load Constants."""
from os.path import expanduser, join

DEV_MAP = {
    "production": "connections",
    "development": "dev-connections",
    "testing": "test-connections",
}

DEFAULT_FILE_LOCATIONS = [
    "./.connections.yaml",

    # the home folder on either Linux or Windows
    join(expanduser("~"), ".connections.yaml")
]



"""Import commonly used functions."""

__version__ = '0.5.7'

from simqle.internal import (
    recordset,
    execute_sql,
    load_connections,
    get_connection,
    reset_connections,
    get_engine,
)
from simqle.helper import bind_sql
from simqle.connection_manager import ConnectionManager

__all__ = [
    "recordset",
    "get_connection",
    "execute_sql",
    "load_connections",
    "ConnectionManager",
    "reset_connections",
    "get_engine",
    "bind_sql",
]




"""Defines the ConnectionManager and Connection Classes."""

import os
import time
import uuid

from yaml import safe_load
from sqlalchemy import create_engine

from urllib.parse import quote_plus
from simqle.constants import DEFAULT_FILE_LOCATIONS, DEV_MAP
from simqle.exceptions import (
    NoConnectionsFileError, UnknownConnectionError,
    MultipleDefaultConnectionsError, EnvironSyncError, UnknownSimqleMode,
    NoDefaultConnectionError,
)
from simqle.helper import bind_sql
from simqle.recordset import RecordSet, RecordScalar, Record
from simqle.logging import logger as log


class ConnectionManager:
    """
    The Connection Manager Class.

    Create an instance of this class with a yaml configuration file. If no
    yaml file is given, the first connection file found in default locations
    will be used instead.

    This is the class from which you execute sql and return recordsets, using
    the public methods self.execute_sql and self.recordset.
    """

    def __init__(self, file_name=None):
        """
        Initialise a ConnectionManager.

        Connections are loaded lazily as required, only the config is loaded
        on initialisation.
        """
        self.connections = {}

        # For backwards compatibility, test mode is given precedence
        if isinstance(os.getenv("SIMQLE_TEST"), str) and os.getenv(
                "SIMQLE_TEST").lower() == "true":

            log.warn("ConnectionManager set to dev mode because SIMQLE_TEST "
                     "is set. SIMQLE_TEST is deprecated, please use "
                     "SIMQLE_MODE instead.")

            self.dev_mode = "testing"
            self.dev_type = "test-connections"

        else:
            self.dev_mode = os.getenv("SIMQLE_MODE", "production")

            log.info(f"ConnectionManager is set to {self.dev_mode} mode")

            self.dev_type = DEV_MAP.get(self.dev_mode)

            if not self.dev_type:
                error_msg = "{} is an unknown simqle mode".format(
                    self.dev_mode)
                raise UnknownSimqleMode(error_msg)

        self._default_connection_name = None
        self.config = None

        if not file_name:
            # file_name isn't given so we search through the possible default
            # file locations, which are in order of priority.
            for default_file_name in DEFAULT_FILE_LOCATIONS:
                try:
                    self.config = self._load_yaml_file(default_file_name)

                    log.info(f"The connections file was loaded from "
                             f"{default_file_name}")
                    break

                except FileNotFoundError:
                    continue

            if not self.config:
                raise NoConnectionsFileError(
                    "No file_name is specified and no files in default "
                    "locations are found.")

        else:
            if isinstance(file_name, dict):
                self.config = file_name

                log.info(f"The connections file was loaded from a dict:"
                         f"{file_name}")
            else:
                self.config = self._load_yaml_file(file_name)

                log.info(f"The connections file was loaded from "
                         f"{file_name}")

        self._check_default_connections()

    # --- Public Methods: ---

    def recordset(self, sql, con_name=None, params=None):
        headings, data = self._recordset(sql, con_name, params=params)
        return RecordSet(headings=headings, data=data)

    def record_scalar(self, sql, con_name=None, params=None):
        headings, data = self._recordset(sql, con_name, params=params)
        return RecordScalar(headings=headings, data=data)

    def record(self, sql, con_name=None, params=None):
        headings, data = self._recordset(sql, con_name, params=params)
        return Record(headings=headings, data=data)

    def execute_sql(self, sql, con_name=None, params=None):
        """Execute SQL on a given connection."""
        execute_id = uuid.uuid4()

        log.info(f"Execution started on {con_name} with id={execute_id}, "
                 f"params={params}, sql={sql}")

        start_time = time.time()
        con_name = self._con_name(con_name)
        connection = self._get_connection(con_name)
        connection.execute_sql(sql, params=params)
        elapsed_time = time.time() - start_time

        log.info(f"Execution id {execute_id} took {elapsed_time:.4f} seconds "
                 f"to complete")

    def get_engine(self, con_name=None):
        """Return the engine of a Connection by it's name."""
        con_name = self._con_name(con_name)
        return self._get_connection(con_name).engine

    def get_connection(self, con_name=None):
        """
        Return the engine of a Connection by it's name.

        Deprecated, only exists for backwards compatibility.
        """
        con_name = self._con_name(con_name)
        return self.get_engine(con_name)  # TODO: add warning

    def reset_connections(self):
        """Remove all current connection objects."""
        self.connections = {}
        self._default_connection_name = None

    # --- Private Methods: ---

    def _recordset(self, sql, con_name=None, params=None):
        """Return headings and data from a connection."""
        recordset_id = uuid.uuid4()

        log.info(f"Query started on {con_name} with id={recordset_id}, "
                 f"params={params}, sql={sql}")

        start_time = time.time()
        con_name = self._con_name(con_name)
        connection = self._get_connection(con_name)
        rst = connection.recordset(sql, params=params)
        elapsed_time = time.time() - start_time

        log.info(f"Query id {recordset_id} took {elapsed_time:.4f} seconds "
                 f"to complete")

        return rst

    @staticmethod
    def _load_yaml_file(connections_file):
        """Load the configuration from the given file."""
        with open(connections_file) as file:
            return safe_load(file.read())

    def _get_connection(self, conn_name):
        """
        Return a connection object from its name.

        Connection objects are created and saved the first time they are
        called.
        """
        # TODO: make conn_name and con_name consistent across the project

        # Return already initialised connection if it exists.
        if conn_name in self.connections:
            return self.connections[conn_name]

        # A new Connection instance is required.
        for conn_config in self.config[self.dev_type]:
            if conn_config["name"] == conn_name:
                self.connections[conn_name] = _Connection(conn_config)
                return self.connections[conn_name]

        raise UnknownConnectionError("Unknown connection {}".format(conn_name))

    def _check_default_connections(self):
        """Check that default settings are set correctly."""
        # See if a default connection exists
        number_of_defaults = 0
        for connection in self.config["connections"]:
            if connection.get("default"):
                number_of_defaults += 1
                self._default_connection_name = connection.get("name")

                log.info(f"Setting the default connection to "
                         f"{connection.get('name')}")

        if not number_of_defaults:
            return

        if number_of_defaults > 1:
            raise MultipleDefaultConnectionsError(
                "More than 1 default connection was specified.")

        if not self.config.get("test-connections"):
            return

        for connection in self.config["test-connections"]:
            if connection.get("default"):
                if connection.get("name") != self._default_connection_name:
                    raise EnvironSyncError("The default connection in "
                                           "connections doesn't match the "
                                           "default connection in the test "
                                           "connections.")

    def _con_name(self, con_name=None):
        if con_name:
            return con_name

        return self._get_default_connection()

    def _get_default_connection(self):
        if not self._default_connection_name:
            raise NoDefaultConnectionError("No Connection name was specified "
                                           "but no default connection exists.")
        return self._default_connection_name


class _Connection:
    """
    The _Connection class.

    This represents a single connection. This class also managers the execution
    of SQL, including the management of transactions.

    The engine is lazily loaded the first time either execute_sql or recordset
    is called.

    This class shouldn't be loaded outside the ConnectionManager class, and so
    is marked as internal only.
    """

    def __init__(self, conn_config):
        """Create a new Connection from a config dict."""
        self.driver = conn_config['driver']
        self._engine = None
        self.name = conn_config['name']

        # Edit the connection based on configuration options.

        # for Microsoft ODBC connections, for example, the connection string
        # must be url escaped. We do this for the user if the url_escape
        # option is True. See here for example and more info:
        # https://docs.sqlalchemy.org/en/13/dialects/mssql.html
        #   #pass-through-exact-pyodbc-string
        if 'url_escape' in conn_config:
            self.connection_string = quote_plus(conn_config['connection'])
        else:
            self.connection_string = conn_config['connection']

        # self.id = uuid.uuid4
        # log.info(f"Connection created for {conn_config['name']} with id "
        #          f"{self.id}")

    def _connect(self):
        """Create an engine based on sqlalchemy's create_engine."""
        self._engine = create_engine(self.driver + self.connection_string)

    @property
    def engine(self):
        """Load the engine if it hasn't been loaded before."""
        if not self._engine:
            self._connect()

        return self._engine

    def execute_sql(self, sql, params=None):
        """Execute :sql: on this connection with named :params:."""
        # action_id = uuid.uuid4
        #
        # log.info("")

        bound_sql = bind_sql(sql, params)

        # TODO: discuss whether a connection should be closed on each
        # transaction.
        connection = self.engine.connect()
        transaction = connection.begin()

        # execute the query, and rollback on error
        try:
            connection.execute(bound_sql)
            transaction.commit()

        except Exception as exception:
            transaction.rollback()
            raise exception

        finally:
            connection.close()

    def recordset(self, sql, params=None):
        """
        Execute <sql> on <con>, with named <params>.

        Return (headings, data)
        """
        # bind the named parameters.
        bound_sql = bind_sql(sql, params)

        # start the connection.
        connection = self.engine.connect()
        transaction = connection.begin()

        # get the results from the query.
        result = connection.execute(bound_sql)
        data = result.fetchall()
        headings = result.keys()

        # commit and close the connection
        transaction.commit()
        connection.close()

        return headings, data
        # return RecordSet(headings=headings, data=data)



from io import open

from setuptools import find_packages, setup

with open("simqle/__init__.py", "r") as f:
    for line in f:
        if line.startswith("__version__"):
            version = line.strip().split("=")[1].strip(' \'"')
            break
    else:
        version = "0.0.1"

with open("README.md", "r", encoding="utf-8") as f:
    readme = f.read()

REQUIRES = ["sqlalchemy", "pyyaml"]
EXTRAS_REQUIRE = {
    "mssqlserver": ["pyodbc"],
    "postgresql": ["psycopg2"],
    "mysql": ["pymysql"],
}

setup(
    name="simqle",
    version=version,
    description="The simple way to SQL",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Tom Malkin",
    author_email="tommalkin28@gmail.com",
    maintainer="Tom Malkin",
    maintainer_email="tommalkin28@gmail.com",
    url="https://github.com/Harlekuin/SimQLe",
    license="MIT",

    keywords=[
        "sql",
    ],

    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
    ],

    install_requires=REQUIRES,
    extras_require=EXTRAS_REQUIRE,
    tests_require=["codecov", "pytest", "coverage", "behave"],

    packages=find_packages(),
)



"""Defines the exceptions used by SimQLe."""


class NoConnectionsFileError(Exception):
    def __init__(self, msg):
        super().__init__(msg)
        self.message = msg


class UnknownConnectionError(Exception):
    def __init__(self, msg):
        super().__init__(msg)
        self.message = msg


class EnvironSyncError(Exception):
    def __init__(self, msg):
        super().__init__(msg)
        self.message = msg


class NoDefaultConnectionError(Exception):
    def __init__(self, msg):
        super().__init__(msg)
        self.message = msg


class MultipleDefaultConnectionsError(Exception):
    def __init__(self, msg):
        super().__init__(msg)
        self.message = msg


class UnknownSimqleMode(Exception):
    def __init__(self, msg):
        super().__init__(msg)
        self.message = msg



"""Internal ConnectionManager."""

from simqle.connection_manager import ConnectionManager

INTERNAL_CONNECTION_MANAGER = None


def load_connections(connections_file="./.connections.yaml"):
    """Load the Internal connection manager."""
    global INTERNAL_CONNECTION_MANAGER

    # if INTERNAL_CONNECTION_MANAGER:
    #     raise Exception("The internal connection manager has already been "
    #                     "loaded, this would overwrite the current one.")
    # TODO: add load details to the ConnectionManager class so we can tell
    # the user at this point what module loaded the internal simqle
    # ConnectionManager instance.

    INTERNAL_CONNECTION_MANAGER = ConnectionManager(file_name=connections_file)


def execute_sql(con_name, sql, params=None):
    """Execute sql on the Internal ConnectionManager."""
    INTERNAL_CONNECTION_MANAGER.execute_sql(sql=sql, con_name=con_name,
                                            params=params)


def recordset(con_name, sql, params=None):
    """Return SQL results from the Internal ConnectionManager."""
    return INTERNAL_CONNECTION_MANAGER.recordset(sql, con_name, params=params)


def get_connection(con_name):
    """Return a connection engine from the Internal ConnectionManager."""
    return INTERNAL_CONNECTION_MANAGER.get_engine(con_name)


def get_engine(con_name):
    """Return a connection engine from the Internal ConnectionManager."""
    return INTERNAL_CONNECTION_MANAGER.get_engine(con_name)


def reset_connections():
    """Clear the internal ConnectionManager's connections."""
    if INTERNAL_CONNECTION_MANAGER:  # TODO: add __bool__ to ConnectionManager
        INTERNAL_CONNECTION_MANAGER.reset_connections()
